import logging
import os
from typing import List

import numpy as np
import pygame
from OpenGL.GL import glClear, GL_COLOR_BUFFER_BIT, glClearColor, glBindVertexArray, \
    glBindBuffer, GL_ARRAY_BUFFER, glCreateShader, \
    GL_VERTEX_SHADER, glShaderSource, glCompileShader, glCreateProgram, glAttachShader, glLinkProgram, glUseProgram, \
    GL_FRAGMENT_SHADER, glPixelStorei, GL_UNSIGNED_BYTE, \
    GL_RGBA, glFlush, glReadPixels, GL_PACK_ALIGNMENT, glEnable, GL_BLEND, glBlendFunc, GL_SRC_ALPHA, \
    GL_ONE_MINUS_SRC_ALPHA, glGetUniformLocation, glUniform4f
from pygame.time import Clock

from gui import Keyboard, GridLayer, CharactersLayer, TextLayer
from gui.GUIFacade import GUIFacade
from gui.Layer import Layer
from gui.pygame import PygameKeyboard


class OpenGLGUIFacade(GUIFacade):

    def __init__(self):
        # Main properties
        self.__screenWidth = 0  # type: int
        self.__screenHeight = 0  # type: int
        self.__clock = None  # type: Clock
        self.__closingRequested = False  # type: bool

        # Shader properties
        self.__shaderProgramId = -1  # type: int
        self.__translationX = 0.0  # type: float
        self.__translationY = 0.0  # type: float

        # Layers
        self.__layers = []  # type: List[Layer]

        # Inputs
        self.__keyboard = PygameKeyboard()

    @property
    def screenWidth(self) -> int:
        return self.__screenWidth

    @property
    def screenHeight(self) -> int:
        return self.__screenHeight

    def createWindow(self, title: str, width: int, height: int):
        os.environ['SDL_VIDEO_CENTERED'] = '1'
        pygame.init()
        pygame.display.set_mode((width, height), pygame.DOUBLEBUF | pygame.OPENGL)
        pygame.display.set_caption(title)
        glClearColor(0.0, 0.0, 0.0, 1.0)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        self.__screenWidth = width
        self.__screenHeight = height

    def createGridLayer(self) -> GridLayer:
        from gui.opengl.OpenGLGridLayer import OpenGLGridLayer
        layer = OpenGLGridLayer(self)
        self.__layers.append(layer)
        return layer

    def createCharactersLayer(self) -> CharactersLayer:
        from gui.opengl.OpenGLCharactersLayer import OpenGLCharactersLayer
        layer = OpenGLCharactersLayer(self)
        self.__layers.append(layer)
        return layer

    def createTextLayer(self) -> TextLayer:
        from gui.opengl.OpenGLTextLayer import OpenGLTextLayer
        layer = OpenGLTextLayer(self)
        self.__layers.append(layer)
        return layer

    def __createShaders(self):
        # Release focus on VBO and VAO
        glBindBuffer(GL_ARRAY_BUFFER, 0)
        glBindVertexArray(0)

        # Create vertex shader
        vertexShaderCode = '''#version 330
        layout (location=0) in vec4 vertex;
        layout (location=1) in vec2 inputUV;
                           out vec2 outputUV;
        uniform vec4 translation;
        void main() {
            gl_Position = vertex + translation;
            outputUV = inputUV;
        }'''
        vertexShaderId = glCreateShader(GL_VERTEX_SHADER)
        glShaderSource(vertexShaderId, vertexShaderCode)
        glCompileShader(vertexShaderId)

        # Create fragment shader
        fragmentShaderCode = '''#version 330
        in vec2 outputUV;
        out vec4 color;
        uniform sampler2D textureColors;
        void main() {
            color = texture(textureColors, outputUV);
        }'''
        fragmentShaderId = glCreateShader(GL_FRAGMENT_SHADER)
        glShaderSource(fragmentShaderId, fragmentShaderCode)
        glCompileShader(fragmentShaderId)

        # Create shader program
        self.__shaderProgramId = glCreateProgram()
        glAttachShader(self.__shaderProgramId, vertexShaderId)
        glAttachShader(self.__shaderProgramId, fragmentShaderId)
        glLinkProgram(self.__shaderProgramId)

    def setTranslation(self, x: float, y: float):
        if len(self.__layers) == 0:
            logging.warning("Can't set the translation because there are no layers")
            return
        layer = self.__layers[0]
        self.__translationX = -x * layer.screenTileWidth
        self.__translationY = y * layer.screenTileHeight

    def init(self):
        self.__createShaders()
        self.__clock = pygame.time.Clock()

    def updateInputs(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.__closingRequested = True
            else:
                self.__keyboard.parsePygameEvent(event)

    def render(self):
        glClear(GL_COLOR_BUFFER_BIT)

        glUseProgram(self.__shaderProgramId)

        translationShaderVar = glGetUniformLocation(self.__shaderProgramId, "translation")
        glUniform4f(translationShaderVar,
                    self.__translationX,
                    self.__translationY,
                    0.0, 0.0)

        for layer in self.__layers:
            layer.draw()

        glUseProgram(0)

        pygame.display.flip()
        self.__clock.tick(60)

    @property
    def keyboard(self) -> Keyboard:
        return self.__keyboard

    @property
    def closingRequested(self) -> bool:
        return self.__closingRequested

    @closingRequested.setter
    def closingRequested(self, b: bool):
        self.__closingRequested = b

    def captureScreenContent(self, x1: int, y1: int, x2: int, y2: int) -> np.ndarray:
        glFlush()
        glPixelStorei(GL_PACK_ALIGNMENT, 4)
        data = glReadPixels(x1, y1, x2, y2, GL_RGBA, GL_UNSIGNED_BYTE)
        capture = np.frombuffer(data, dtype=np.uint8)
        capture = capture.reshape(y2 - y1, x2 - x1, 4)
        capture = np.flip(capture, axis=0)
        return capture

    def quit(self):
        for layer in self.__layers:
            layer.deleteObjects()
        pygame.quit()
        quit()
