import os
from typing import Union, Dict

import numpy as np

from gui import TextLayer
from gui.opengl import OpenGLMesh, OpenGLGUIFacade
from gui.pygame import TextTileset, CharacterTile


class OpenGLTextLayer(OpenGLMesh, TextLayer):

    def __init__(self, gui: OpenGLGUIFacade):
        super(OpenGLTextLayer, self).__init__(gui)
        self.__vertices = None  # type: Union[None, np.ndarray]
        self.__faces = None  # type: Union[None, np.ndarray]
        self.__uvMap = None  # type: Union[None, np.ndarray]
        self.__tileset = TextTileset()
        self.__transparentTile = CharacterTile()

    @property
    def maxCharacterCount(self) -> int:
        return self.__vertices.shape[0]

    def setFont(self, fontName: str, tileHeight: int, color: (int, int, int)):
        characters = \
            " !\"#$%&'()*+,-./" \
            "0123456789:;<=>?" \
            "@ABCDEFGHIJKLMNO" \
            "PQRSTUVWXYZ[\\]^_" \
            "`abcdefghijklmno" \
            "pqrstuvwxyz{|}~"
        self.__tileset.setFont(fontName, tileHeight, color)
        self.__updateTileset(characters)

    def setMaxCharacterCount(self, maxCount: int):
        if self.__vertices is not None:
            raise RuntimeError("Update of max character count is not supported")

        self.__vertices = np.empty([maxCount, 4, 2], dtype=np.float32)
        self.__uvMap = np.empty([maxCount, 4, 2], dtype=np.float32)
        self.__faces = np.empty([maxCount, 4], dtype=np.uint)
        faceCount = 0
        for quadIndex in range(maxCount):
            self.__setQuadLocation(quadIndex, 0, 0, 0, 0)
            self.__setQuadTile(
                quadIndex,
                self.__transparentTile.x1, self.__transparentTile.y1,
                self.__transparentTile.x2, self.__transparentTile.y2
            )

            self.__faces[quadIndex, 0] = faceCount * 4
            self.__faces[quadIndex, 1] = faceCount * 4 + 1
            self.__faces[quadIndex, 2] = faceCount * 4 + 2
            self.__faces[quadIndex, 3] = faceCount * 4 + 3
            faceCount += 1

        self.setData(self.__vertices, self.__faces, self.__uvMap, dynamic=True)

    def draw(self):
        self.updateData(self.__vertices, self.__faces, self.__uvMap)
        super(OpenGLTextLayer, self).draw()

    def __updateTileset(self, characters: str):
        # Add characters to the tileset
        if self.__tileset.addCharacters(characters):
            # The transparent could have changed
            self.__transparentTile = self.__tileset.getCharacterTile(" ")

            # Send the tileset to the GPU
            self.setTileset(
                self.__tileset.toNumpyArray(),
                self.__tileset.tileDefaultWidth,
                self.__tileset.tileDefaultHeight
            )

            if False:  # Save generated tileset (to debug/illustrate)
                from PIL import Image
                if not hasattr(self, "tilesetUpdates"):
                    self.tilesetUpdates = 0
                self.tilesetUpdates += 1
                Image.fromarray(self.__tileset.toNumpyArray()).save("charactersTileset{}.png".format(self.tilesetUpdates))

    def setText(self, x: float, y: float, content: str):
        if len(content) > self.maxCharacterCount:
            content = content[0:self.maxCharacterCount]

        # Update the tileset (only for new characters)
        self.__updateTileset(content)

        x0 = x
        # Characters to display
        for charIndex, character in enumerate(content):
            if character == "\n":
                x = x0
                y += self.tileHeight
                continue
            tile = self.__tileset.getCharacterTile(character)
            self.__setQuadTile(charIndex, tile.x1, tile.y1, tile.x2, tile.y2)
            self.__setQuadLocation(charIndex, x, y, x + tile.width, y + tile.height)
            x += tile.width

        # Remaining characters are transparent
        for charIndex in range(len(content), self.maxCharacterCount):
            self.__setQuadTile(charIndex,
                self.__transparentTile.x1, self.__transparentTile.y1,
                self.__transparentTile.x2, self.__transparentTile.y2
            )

    def __setQuadLocation(self, quadIndex: int, x1: float, y1: float, x2: float, y2: float):
        assert 0 <= quadIndex < self.__vertices.shape[0]

        spriteScreenX1 = -1 + x1 * self.screenPixelWidth
        spriteScreenY1 = 1 - y1 * self.screenPixelHeight
        spriteScreenX2 = -1 + x2 * self.screenPixelWidth
        spriteScreenY2 = 1 - y2 * self.screenPixelHeight
        self.__vertices[quadIndex, 0] = [spriteScreenX1, spriteScreenY2]
        self.__vertices[quadIndex, 1] = [spriteScreenX1, spriteScreenY1]
        self.__vertices[quadIndex, 2] = [spriteScreenX2, spriteScreenY1]
        self.__vertices[quadIndex, 3] = [spriteScreenX2, spriteScreenY2]

    def __setQuadTile(self, quadIndex: int, tileX1: float, tileY1: float, tileX2: float, tileY2: float):
        assert 0 <= quadIndex < self.__vertices.shape[0]

        spriteTextureX1 = tileX1 / self.textureWidth
        spriteTextureY1 = tileY1 / self.textureHeight
        spriteTextureX2 = tileX2 / self.textureWidth
        spriteTextureY2 = tileY2 / self.textureHeight
        self.__uvMap[quadIndex, 0] = [spriteTextureX1, spriteTextureY2]
        self.__uvMap[quadIndex, 1] = [spriteTextureX1, spriteTextureY1]
        self.__uvMap[quadIndex, 2] = [spriteTextureX2, spriteTextureY1]
        self.__uvMap[quadIndex, 3] = [spriteTextureX2, spriteTextureY2]
