import logging
from typing import Union, Dict

import numpy as np
import pygame


class CharacterTile:
    __slots__ = ['x1', 'y1', 'x2', 'y2']

    def __init__(self, x1: int=0, y1: int=0, x2: int=0, y2: int=0):
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2

    @property
    def width(self) -> int:
        return self.x2 - self.x1

    @property
    def height(self) -> int:
        return self.y2 - self.y1


class TextTileset:
    def __init__(self):
        self.__font = None  # type: Union[None, pygame.font.Font]
        self.__color = (255, 255, 255)  # type: (int, int, int)
        self.__tileset = None  # type: Union[None, pygame.Surface]
        self.__nextTileX = 0  # type: int
        self.__nextTileY = 0  # type: int
        self.__tileDefaultWidth = 0  # type: int
        self.__tileDefaultHeight = 0  # type: int
        self.__characterTiles = {}  # type: Dict[str, CharacterTile]

    @property
    def tileDefaultWidth(self) -> int:
        return self.__tileDefaultWidth

    @property
    def tileDefaultHeight(self) -> int:
        return self.__tileDefaultHeight

    def __resize(self, width: int, height: int):

        # GPUs prefer square image size with a power of two
        def nextPowerOfTwo(x: int) -> int:
            return 1 if x == 0 else 2 ** (x - 1).bit_length()
        imageWidth = nextPowerOfTwo(width)
        imageHeight = nextPowerOfTwo(height)
        imageSize = max(imageWidth, imageHeight)
        self.__tileset = pygame.Surface((imageSize, imageSize), flags=pygame.SRCALPHA)

        # Copy current characters in the new tileset
        currentCharacters = "".join(self.__characterTiles.keys())
        self.__characterTiles.clear()
        self.__nextTileX = 0
        self.__nextTileY = 0
        self.addCharacters(currentCharacters)

    def setFont(self, fontName: str, tileHeight: int, color: (int, int, int)):
        # Render a single character to find the required tile height
        fontSize = int(tileHeight * 1.3)
        for iterations in range(100):  # Up to 100 tries
            font = pygame.font.Font(fontName, fontSize)
            surface = font.render("M", False, color)
            if surface.get_height() == tileHeight:
                break
            elif surface.get_height() > tileHeight:
                fontSize -= 1
            else:
                fontSize += 1
        self.__font = font
        self.__color = color
        self.__tileDefaultWidth = surface.get_width()
        self.__tileDefaultHeight = surface.get_height()
        self.__characterTiles = {}
        self.__resize(256, 256)

    def addCharacters(self, characters: str) -> bool:
        tilesetUpdated = False
        for character in characters:
            if character in self.__characterTiles:
                continue
            if character == "\n":
                continue
            tilesetUpdated = True

            # Render the character
            surface = self.__font.render(character, False, self.__color)
            characterWidth = surface.get_width()
            if self.__tileDefaultHeight < surface.get_height():
                logging.warning("The height of character {} is higher than the default.".format(character))

            # Ensure that next tile location in inside the tileset
            if (self.__nextTileX + characterWidth) > self.__tileset.get_width():
                self.__nextTileX = 0
                self.__nextTileY += self.__tileDefaultHeight
            if (self.__nextTileY + self.__tileDefaultHeight) > self.__tileset.get_height():
                self.__resize(2 * self.__tileset.get_width(), 2 * self.__tileset.get_height())
                if (self.__nextTileX + characterWidth) > self.__tileset.get_width():
                    self.__nextTileX = 0
                    self.__nextTileY += self.__tileDefaultHeight

            # Blit the character in the tileset
            self.__tileset.blit(surface, (self.__nextTileX, self.__nextTileY))

            # Compute and save the tile location
            self.__characterTiles[character] = CharacterTile(
                self.__nextTileX, self.__nextTileY,
                self.__nextTileX + characterWidth, self.__nextTileY + self.__tileDefaultHeight
            )

            # Update tile next location
            self.__nextTileX += characterWidth
        return tilesetUpdated

    def toNumpyArray(self) -> np.ndarray:
        imageArray = np.zeros((self.__tileset.get_width(), self.__tileset.get_height(), 4), dtype=np.uint8)
        imageArray[..., 0] = pygame.surfarray.pixels_red(self.__tileset)
        imageArray[..., 1] = pygame.surfarray.pixels_green(self.__tileset)
        imageArray[..., 2] = pygame.surfarray.pixels_blue(self.__tileset)
        imageArray[..., 3] = pygame.surfarray.pixels_alpha(self.__tileset)
        return imageArray.transpose((1, 0, 2))

    def getCharacterTile(self, character: str) -> CharacterTile:
        return self.__characterTiles[character]

